function [mdl1, mdl2, bestPwrs1, bestPwrs2] = bestrathet2(XlFile,outFile)

warning('off');
format long

dt = datetime('now','TimeZone','local','Format','y-MMM-d HH_mm_ss');
dtstr = string(dt);
if nargin < 2, outFile = strcat("bestrathet2_",dtstr,".txt"); end
if nargin < 1, XlFile = strcat(pwd,'\bestrathet2.xlsx'); end

diary(outFile)

fprintf("Function bestrathet2\n\n");
fprintf("%s\n", dtstr);
fprintf("Please wait ....\n\n");

fprintf("Excel file used is %s\n", XlFile);
fprintf("Diary file used is %s\n", outFile);
% Open Excel as a COM Automation server
Excel = actxserver('Excel.Application');
% Open Excel workbook
Workbook = Excel.Workbooks.Open(XlFile);
% Clear the content of the sheet
Workbook.Worksheets.Item('List').Range('A2:Z100000').ClearContents;  
Workbook.Worksheets.Item('List2').Range('A2:Z100000').ClearContents;  
Workbook.Worksheets.Item('Project').Range('A2:Z100000').ClearContents;  
Workbook.Worksheets.Item('Project2').Range('A2:Z100000').ClearContents;  
% Now save/close/quit
Workbook.Save;
Excel.Workbook.Close;
invoke(Excel, 'Quit');

hdrList = {'Adj R-square','YPwr','X1Pwr','YX1Pwr','Incercept','X1','YX1'};
writecell(hdrList,XlFile,'Sheet','List');
writecell(hdrList,XlFile,'Sheet','List2');
hdrProject = {'Y','X1','YCalc','%Err'};
writecell(hdrProject,XlFile,'Sheet','Project');
writecell(hdrProject,XlFile,'Sheet','Project2');

xyData = readmatrix(XlFile,'Sheet','Data');
[maxrows,~] = size(xyData);
ydata = xyData(:,1);
xdata = xyData(:,2);
xyTransf = readmatrix(XlFile,'Sheet','Transf');
yPwrs = xyTransf(1:3,1);
x1Pwrs = xyTransf(1:3,2);
x2Pwrs = xyTransf(1:3,3);
minAdjR2 = xyTransf(1,4);
yBuffer = zeros(maxrows,1);
x1Buffer = zeros(maxrows,1);
x2Buffer = zeros(maxrows,1);
yCols = 0;
for yPwr =yPwrs(1):yPwrs(2):yPwrs(3)
  yCols = yCols + 1;
  yBuffer(:,yCols)= fx(ydata,yPwr);
end

x1Cols = 0;
for x1Pwr =x1Pwrs(1):x1Pwrs(2):x1Pwrs(3)
  x1Cols = x1Cols + 1;
  x1Buffer(:,x1Cols)= fx(xdata,x1Pwr);
end

x2Cols = 0;
for x2Pwr =x2Pwrs(1):x2Pwrs(2):x2Pwrs(3)
  x2Cols = x2Cols + 1;
  x2Buffer(:,x2Cols)= fx(xdata,x2Pwr);
end

bestAdjR2 = 0;
gRow = 0;
xyList = zeros(1,7);

yCols = 0;
for yPwr =yPwrs(1):yPwrs(2):yPwrs(3)
  yCols = yCols + 1;
  y = yBuffer(:,yCols);

  x1Cols = 0 ; 
  for x1Pwr =x1Pwrs(1):x1Pwrs(2):x1Pwrs(3)
    x1Cols = x1Cols + 1;
    x1 = x1Buffer(:,x1Cols);

    x2Cols = 0  ;
    for x2Pwr =x2Pwrs(1):x2Pwrs(2):x2Pwrs(3)
      x2Cols = x2Cols + 1;
      x2 = x2Buffer(:,x2Cols);

      x2 = -1 .* y .* x2;
      X = [x1 x2];
    
      mdl = fitlm(X,y);    
      
      if mdl.Rsquared.Adjusted >= bestAdjR2
        bestAdjR2 = mdl.Rsquared.Adjusted;
        bestYpwr = yPwr;
        bestX1pwr = x1Pwr;
        bestX2pwr = x2Pwr;
        bestmdl = mdl;
      end

      if mdl.Rsquared.Adjusted >= minAdjR2
        gRow = gRow + 1;
        xyList(gRow,1) = mdl.Rsquared.Adjusted;
        xyList(gRow,2) = yPwr;
        xyList(gRow,3) = x1Pwr;
        xyList(gRow,4) = x2Pwr;
        xyList(gRow,5)= mdl.Coefficients{1,1};
        xyList(gRow,6)= mdl.Coefficients{2,1};
        xyList(gRow,7)= mdl.Coefficients{3,1};
      end
    end
  end
end
mdl = bestmdl;
bestPwrs1 = [bestYpwr bestX1pwr bestX2pwr];
bestPwrs2 = [bestYpwr bestX1pwr bestX2pwr]; % default assignement
[numRes,~] = size(xyList);
xyList = sortrows(xyList,1,"descend");
writematrix(xyList, XlFile, 'Sheet', 'List', 'Range',strcat('A2:G',num2str(numRes+1)));

% calculate projecttions
a = mdl.Coefficients{1,1};
b = mdl.Coefficients{2,1};
c = mdl.Coefficients{3,1};

ycalc = (a + b.*fx(xdata,bestX1pwr))./(1 + c.*fx(xdata,bestX2pwr));
if bestYpwr > 0
  ycalc = ycalc.^(1/bestYpwr);
elseif bestYpwr < 0
  ycalc = ycalc.^abs(bestYpwr);
else
  ycalc = exp(ycalc);
end
percErr = 100.*(ycalc-ydata)./ydata;
ProjMat = [ydata xdata ycalc percErr];
writematrix(ProjMat, XlFile, 'Sheet', 'Project', 'Range', strcat("A2:Z", num2str(1+length(ydata))));

% Print the summary of the best regression results
fprintf("Summary:\n");
fprintf("Adj R-sqr = %14.12f", bestAdjR2);
mdl
mdl1 = mdl;
mdl2 = mdl; % default assigment
fprintf("Best Y power = %f\n", bestYpwr);
fprintf("Best numerator X power = %f\n", bestX1pwr);
fprintf("Best denominator X power = %f\n", bestX2pwr);

if bestYpwr ~= 1
  fprintf("=============================================\n");
  fprintf("\n\n------------------ Phase 2 -----------------\n\n");
  fprintf("=============================================\n");

  bestAdjR2 = 0;
  gRow = 0;
  xyList = zeros(1,7);

  y = ydata;

  x1Cols = 0 ;
  for x1Pwr =x1Pwrs(1):x1Pwrs(2):x1Pwrs(3)
    x1Cols = x1Cols + 1;
    x1 = x1Buffer(:,x1Cols);

    x2Cols = 0  ;
    for x2Pwr =x2Pwrs(1):x2Pwrs(2):x2Pwrs(3)
      x2Cols = x2Cols + 1;
      x2 = x2Buffer(:,x2Cols);

      x2 = -1 .* y .* x2;
      X = [x1 x2];

      mdl = fitlm(X,y);

      if mdl.Rsquared.Adjusted >= bestAdjR2
        bestAdjR2 = mdl.Rsquared.Adjusted;
        bestYpwr = 1;
        bestX1pwr = x1Pwr;
        bestX2pwr = x2Pwr;
        bestmdl = mdl;
      end

      if mdl.Rsquared.Adjusted >= minAdjR2
        gRow = gRow + 1;
        xyList(gRow,1) = mdl.Rsquared.Adjusted;
        xyList(gRow,2) = 1;
        xyList(gRow,3) = x1Pwr;
        xyList(gRow,4) = x2Pwr;
        xyList(gRow,5)= mdl.Coefficients{1,1};
        xyList(gRow,6)= mdl.Coefficients{2,1};
        xyList(gRow,7)= mdl.Coefficients{3,1};
      end
    end
  end
  mdl = bestmdl;
  bestPwrs2 = [bestYpwr bestX1pwr bestX2pwr];
  [numRes,~] = size(xyList);
  xyList = sortrows(xyList,1,"descend");
  writematrix(xyList, XlFile, 'Sheet', 'List2', 'Range',strcat('A2:G',num2str(numRes+1)));
  
  % calculate projecttions
  a = mdl.Coefficients{1,1};
  b = mdl.Coefficients{2,1};
  c = mdl.Coefficients{3,1};
  
  ycalc = (a + b.*fx(xdata,bestX1pwr))./(1 + c.*fx(xdata,bestX2pwr));
  percErr = 100.*(ycalc-ydata)./ydata;
  ProjMat = [ydata xdata ycalc percErr];
  writematrix(ProjMat, XlFile, 'Sheet', 'Project2', 'Range', strcat("A2:Z", num2str(1+length(ydata))));
  
  % Print the summary of the best regression results
  fprintf("Summary:\n");
  fprintf("Adj R-sqr = %14.12f", bestAdjR2);
 
  mdl
  mdl2 = mdl;
  fprintf("Best Y power = %f\n", bestYpwr);
  fprintf("Best numerator X power = %f\n", bestX1pwr);
  fprintf("Best denominator X power = %f\n", bestX2pwr);
end

format short
diary off
end

function y = fx(x,pwr)
  if pwr > 0
    y = x.^pwr;
  elseif pwr < 0
    y = 1./x.^abs(pwr);
  else
    y = log(x);
  end
end