function [mdl1, mdl2, bestPwrs1, bestPwrs2] = bestrathet8Flex(XlFile,outFile)

warning('off')
format long

dt = datetime('now','TimeZone','local','Format','y-MMM-d HH_mm_ss');
dtstr = string(dt);
if nargin < 2, outFile = strcat("bestrathet8Flex1_",dtstr,".txt"); end
if nargin < 1, XlFile = strcat(pwd,'\bestrathet8Flex1.xlsx'); end
diary(outFile)
% Flexible rational heteronomials
fprintf("Function bestrathet8Flex\n\n");
fprintf("%s\n", dtstr);
fprintf("Please wait ....\n\n");

fprintf("Excel file used is %s\n", XlFile);
fprintf("Diary file is %s\n", outFile);

% Open Excel as a COM Automation server
Excel = actxserver('Excel.Application');
% Open Excel workbook
Workbook = Excel.Workbooks.Open(XlFile);
% Clear the content of the sheets
Workbook.Worksheets.Item('List').Range('A2:Z100000').ClearContents;
Workbook.Worksheets.Item('List2').Range('A2:Z100000').ClearContents;
Workbook.Worksheets.Item('Project').Range('A2:Z100000').ClearContents;
Workbook.Worksheets.Item('Project2').Range('A2:Z100000').ClearContents;
% Now save/close/quit
Workbook.Save;
Excel.Workbook.Close;
invoke(Excel, 'Quit');

xyData = readmatrix(XlFile,'Sheet','Data');
[maxrows,cols] = size(xyData);
ydata = xyData(:,1);
xData = xyData(:,2:cols);
xyTransf = readmatrix(XlFile,'Sheet','Transf');
nTerms = xyTransf(1,1);
xidx = xyTransf(1,2:nTerms+1);
flag = zeros(nTerms,1);

hdrList = {'Adj R-square','YPwr'};
c = 3;
for i=1:nTerms
   j = abs(xidx(i));
   if xidx(i) > 0
     s = strcat("XPwr", num2str(j)); 
   else 
     s = strcat("YXPwr", num2str(j));
   end
   hdrList(c) = cellstr(s);
   c = c + 1;
end
   
hdrList(c) = {'Intercept'};
c = c + 1;
for i=1:nTerms
   j = abs(xidx(i));
   if xidx(i) > 0
     s = strcat("XPwr", num2str(j)); 
   else 
     s = strcat("YXPwr", num2str(j));
   end
   hdrList(c) = cellstr(s);
   c = c + 1;
end

writecell(hdrList,XlFile,'Sheet','List');
writecell(hdrList,XlFile,'Sheet','List2');

hdrProject = {'Y'};
c = 1;
for i=1:nTerms
   j = abs(xidx(i));
   if xidx(i) > 0
     s = strcat("X", num2str(j)); 
     c = c + 1;
     hdrProject(c) = cellstr(s);
   end 
end   
c = c + 1;
hdrProject(c) = {'YCalc'};
c = c + 1;
hdrProject(c) = {'%Err'};
writecell(hdrProject,XlFile,'Sheet','Project');
writecell(hdrProject,XlFile,'Sheet','Project2');

flag = zeros(nTerms,1);
for i=1:nTerms
  if xidx(i) < 0, flag(i) = 1; end
  xidx(i) = abs(xidx(i));
end

yPwrs = xyTransf(2:4,1);
x1Pwrs = xyTransf(2:4,2);
x2Pwrs = xyTransf(2:4,3);
x3Pwrs = xyTransf(2:4,4);
if nTerms >= 4
  x4Pwrs = xyTransf(2:4,5);
else
  x4Pwrs = [1 1 1];
end
if nTerms >= 5
  x5Pwrs = xyTransf(2:4,6);
else
  x5Pwrs = [1 1 1];
end
if nTerms >= 6
  x6Pwrs = xyTransf(2:4,7);
else
  x6Pwrs = [1 1 1];
end
if nTerms >= 7
  x7Pwrs = xyTransf(2:4,8);
else
  x7Pwrs = [1 1 1];
end
if nTerms >= 8
  x8Pwrs = xyTransf(2:4,9);
else
  x8Pwrs = [1 1 1];
end

minAdjR2 = xyTransf(1,nTerms+2);
yBuffer = zeros(maxrows,1);
x1Buffer = zeros(maxrows,1);
x2Buffer = zeros(maxrows,1);
x3Buffer = zeros(maxrows,1);
x4Buffer = zeros(maxrows,1);
x5Buffer = zeros(maxrows,1);
x6Buffer = zeros(maxrows,1);
x7Buffer = zeros(maxrows,1);
x8Buffer = zeros(maxrows,1);

yCols = 0;
for yPwr =yPwrs(1):yPwrs(2):yPwrs(3)
  yCols = yCols + 1;
  yBuffer(:,yCols) = fx(ydata,yPwr);
end

x1Cols = 0;
j = xidx(1);
for x1Pwr =x1Pwrs(1):x1Pwrs(2):x1Pwrs(3)
  x1Cols = x1Cols + 1;
  x1Buffer(:,x1Cols) = fx(xData(:,j),x1Pwr);
end

x2Cols = 0;
j = xidx(2);
for x2Pwr =x2Pwrs(1):x2Pwrs(2):x2Pwrs(3)
  x2Cols = x2Cols + 1;
  x2Buffer(:,x2Cols) = fx(xData(:,j),x2Pwr);
end

x3Cols = 0;
j = xidx(3);
for x3Pwr =x3Pwrs(1):x3Pwrs(2):x3Pwrs(3)
  x3Cols = x3Cols + 1;
  x3Buffer(:,x3Cols) = fx(xData(:,j),x3Pwr);
end

if nTerms >= 4
  x4Cols = 0;
  j = xidx(4);
  for x4Pwr =x4Pwrs(1):x4Pwrs(2):x4Pwrs(3)
    x4Cols = x4Cols + 1;
    x4Buffer(:,x4Cols) = fx(xData(:,j),x4Pwr);
  end
end

if nTerms >= 5
  x5Cols = 0;
  j = xidx(5);
  for x5Pwr =x5Pwrs(1):x5Pwrs(2):x5Pwrs(3)
    x5Cols = x5Cols + 1;
    x5Buffer(:,x5Cols) = fx(xData(:,j),x5Pwr);
  end
end

if nTerms >= 6
  x6Cols = 0;
  j = xidx(6);
  for x6Pwr =x6Pwrs(1):x6Pwrs(2):x6Pwrs(3)
    x6Cols = x6Cols + 1;
    x6Buffer(:,x6Cols) = fx(xData(:,j),x6Pwr);
  end
end

if nTerms >= 7
  x7Cols = 0;
  j = xidx(7);
  for x7Pwr =x7Pwrs(1):x7Pwrs(2):x7Pwrs(3)
    x7Cols = x7Cols + 1;
    x7Buffer(:,x7Cols) = fx(xData(:,j),x7Pwr);
  end
end

if nTerms >= 8
  x8Cols = 0;
  j = xidx(8);
  for x8Pwr =x8Pwrs(1):x8Pwrs(2):x8Pwrs(3)
    x8Cols = x8Cols + 1;
    x8Buffer(:,x8Cols)= fx(xData(:,j),x8Pwr);
  end
end

for iRun=1:2

  bestAdjR2 = 0;
  gRow = 0;
  xyList = zeros(1,1+2*(1+nTerms));
  yCols = 0;
  for yPwr =yPwrs(1):yPwrs(2):yPwrs(3)
    yCols = yCols + 1;
    y = yBuffer(:,yCols);
   
    x1Cols = 0 ;
    for x1Pwr =x1Pwrs(1):x1Pwrs(2):x1Pwrs(3)
      x1Cols = x1Cols + 1;
      x1 = x1Buffer(:,x1Cols);

      x2Cols = 0  ;
      for x2Pwr =x2Pwrs(1):x2Pwrs(2):x2Pwrs(3)
        x2Cols = x2Cols + 1;
        x2 = x2Buffer(:,x2Cols);

        if flag(2)==1, x2 = -1 .* y .* x2; end

        x3Cols = 0  ;
        for x3Pwr =x3Pwrs(1):x3Pwrs(2):x3Pwrs(3)
          x3Cols = x3Cols + 1;
          x3 = x3Buffer(:,x3Cols);

          if flag(3)==1, x3 = -1 .* y .* x3; end

          x4Cols = 0 ;
          for x4Pwr =x4Pwrs(1):x4Pwrs(2):x4Pwrs(3)
            if nTerms >= 4
              x4Cols = x4Cols + 1;
              x4 = x4Buffer(:,x4Cols);

              if flag(4)==1, x4 = -1 .* y .* x4; end
            else
              x4 = [];
            end

            x5Cols = 0;
            for x5Pwr =x5Pwrs(1):x5Pwrs(2):x5Pwrs(3)
              if nTerms >= 5
                x5Cols = x5Cols + 1;
                x5 = x5Buffer(:,x5Cols);

                if flag(5)==1, x5 = -1 .* y .* x5; end
              else
                x5 = [];
              end

                x6Cols = 0;
                for x6Pwr =x6Pwrs(1):x6Pwrs(2):x6Pwrs(3)
                  if nTerms >= 6
                    x6Cols = x6Cols + 1;
                    x6 = x6Buffer(:,x6Cols);

                    if flag(6)==1, x6 = -1 .* y .* x6; end
                  else
                    x6 = [];
                  end

                  x7Cols = 0;
                  for x7Pwr =x7Pwrs(1):x7Pwrs(2):x7Pwrs(3)
                    if nTerms >= 7
                      x7Cols = x7Cols + 1;
                      x7 = x7Buffer(:,x7Cols);

                      if flag(7)==1, x7 = -1 .* y .* x7; end
                    else
                      x7 = [];
                    end

                    x8Cols = 0  ;
                    for x8Pwr =x8Pwrs(1):x8Pwrs(2):x8Pwrs(3)
                      if nTerms >= 8
                        x8Cols = x8Cols + 1;
                        x8 = x8Buffer(:,x8Cols);

                        if flag(8)==1, x8 = -1 .* y .* x8; end
                      else
                        x8 = [];
                      end

                      X = [x1 x2 x3 x4 x5 x6 x7 x8];

                      mdl = fitlm(X,y);

                      if mdl.Rsquared.Adjusted >= bestAdjR2
                        bestAdjR2 = mdl.Rsquared.Adjusted;
                        bestYpwr = yPwr;
                        bestx1pwr = x1Pwr;
                        bestx2pwr = x2Pwr;
                        bestx3pwr = x3Pwr;
                        if nTerms >=4, bestx4pwr = x4Pwr; end
                        if nTerms >=5, bestx5pwr = x5Pwr; end
                        if nTerms >=6, bestx6pwr = x6Pwr; end
                        if nTerms >=7,bestx7pwr = x7Pwr; end
                        if nTerms >=8, bestx8pwr = x8Pwr; end
                        bestmdl = mdl;
                      end

                      if mdl.Rsquared.Adjusted >= minAdjR2
                        gRow = gRow + 1;
                        xyList(gRow,1) = mdl.Rsquared.Adjusted;
                        xyList(gRow,2) = yPwr;
                        xyList(gRow,3) = x1Pwr;
                        xyList(gRow,4) = x2Pwr;
                        xyList(gRow,5) = x3Pwr;
                        gCol = 6;
                        if nTerms >= 4
                          xyList(gRow,gCol) = x4Pwr;
                          gCol = gCol + 1;
                        end
                        if nTerms >= 5
                          xyList(gRow,gCol) = x5Pwr;
                          gCol = gCol + 1;
                        end
                        if nTerms >= 6
                          xyList(gRow,gCol) = x6Pwr;
                          gCol = gCol + 1;
                        end
                        if nTerms >= 7
                          xyList(gRow,gCol) = x7Pwr;
                          gCol = gCol + 1;
                        end
                        if nTerms >= 8
                          xyList(gRow,gCol) = x8Pwr;
                          gCol = gCol + 1;
                        end
                        xyList(gRow,gCol)= mdl.Coefficients{1,1};
                        gCol = gCol + 1;
                        xyList(gRow,gCol)= mdl.Coefficients{2,1};
                        gCol = gCol + 1;
                        xyList(gRow,gCol)= mdl.Coefficients{3,1};
                        gCol = gCol + 1;
                        xyList(gRow,gCol)= mdl.Coefficients{4,1};
                        gCol = 5;
                        if nTerms >= 4
                          xyList(gRow,gCol)= mdl.Coefficients{gCol,1};
                          gCol = gCol + 1;
                        end
                        if nTerms >= 5
                          xyList(gRow,gCol)= mdl.Coefficients{gCol,1};
                          gCol = gCol + 1;
                        end
                        if nTerms >= 6
                          xyList(gRow,gCol)= mdl.Coefficients{gCol7,1};
                          gCol = gCol + 1;
                        end
                        if nTerms >= 7
                          xyList(gRow,gCol)= mdl.Coefficients{gCol,1};
                          gCol = gCol + 1;
                        end
                        if nTerms >= 8
                          xyList(gRow,gCol)= mdl.Coefficients{gCol,1};
                        end
                      end
                    end
                  end
                end
              end
            end
          end
        end
      end
    end

  [numRes,~] = size(xyList);
  mdl = bestmdl;
  xyList = sortrows(xyList,1,"descend");
  if iRun==1
    writematrix(xyList, XlFile, 'Sheet', 'List', 'Range',strcat('A2:S',num2str(numRes+1)));
  elseif iRun==2 && bestYpwr ~= 1
    writematrix(xyList, XlFile, 'Sheet', 'List2', 'Range',strcat('A2:S',num2str(numRes+1)));
  end
  % calculate projected values
  ycalc = ydata;
  for i=1:length(ydata)
    sum1 = mdl.Coefficients{1,1};
    sum2 = 1;
    for j=1:nTerms
      z =  mdl.Coefficients{j+1,1}*fx(xData(i,xidx(j)),xyList(1,j+2));
      if flag(j) == 0
        sum1 = sum1 + z;
      else
        sum2 = sum2 + z;
      end
    end

    ycalc(i) = sum1 / sum2;
    if iRun==1
      if bestYpwr > 0
        ycalc(i) = ycalc(i)^(1/bestYpwr);
      elseif bestYpwr < 0
        ycalc(i) = ycalc(i)^abs(bestYpwr);
      else
        ycalc(i) = exp(ycalc(i));
      end
    end
  end
  percErr = 100.*(ycalc-ydata)./ydata;
  ProjMat = [ydata xData ycalc percErr];
  if iRun==1
    writematrix(ProjMat, XlFile, 'Sheet', 'Project', 'Range', strcat("A2:Z", num2str(length(ydata))));
  elseif iRun==2 && bestYpwr ~= 1
    writematrix(ProjMat, XlFile, 'Sheet', 'Project2', 'Range', strcat("A2:Z", num2str(length(ydata))));
  end
  % Print the summary of the best regression results
  if iRun==1, mdl1=mdl; end
  if iRun==1 || (iRun==2 && bestYpwr ~= 1)
    fprintf("Summary:\n");
    mdl

    fprintf("Best Y power = %f\n", bestYpwr);
    fprintf("Best numerator X%d power = %f\n", xidx(1), bestx1pwr);
    fprintf("Best %s X%d power = %f\n", classify(flag(2)), xidx(2), bestx2pwr);
    fprintf("Best %s X%d power = %f\n", classify(flag(3)), xidx(3), bestx3pwr);
    if nTerms >= 4
      fprintf("Best %s X%d power = %f\n", classify(flag(4)), xidx(4), bestx4pwr);
    end
    if nTerms >= 5
      fprintf("Best %s X%d power = %f\n", classify(flag(5)), xidx(5), bestx5pwr);
    end
    if nTerms >= 6
      fprintf("Best %s X%d power = %f\n", classify(flag(6)), xidx(6), bestx6pwr);
    end
    if nTerms >= 7
      fprintf("Best %s X%d power = %f\n", classify(flag(7)), xidx(7), bestx7pwr);
    end
    if nTerms >= 8
      fprintf("Best %s X%d power = %f\n", classify(flag(8)), xidx(8), bestx8pwr);
    end
  end
  
  if iRun==1 && bestYpwr ~= 1
    fprintf("=============================================\n");
    fprintf("\n\n------------------ Phase 2 -----------------\n\n");
    fprintf("------------------- Set yPwr = 1 -------------------\n");
    fprintf("=============================================\n");
    yPwrs = [1 1 1];
  end
end % iRun loop

diary off
end

function y = fx(x,pwr)
  if pwr > 0
    y = x.^pwr;
  elseif pwr < 0
    y = 1./x.^abs(pwr);
  else
    y = log(x);
  end
end

function s=classify(f)
  if f==0
    s = "numerator";
  else
    s = "denominator";
  end
end